import argparse
import collections
import copy
import csv
import json
import math
import os
import re
import statistics
from pathlib import Path
from statistics import stdev
from typing import Dict, Tuple, Optional, List, Iterable, Set

from randovania.game_description.default_database import game_description_for
from randovania.game_description.world.node import PickupNode, LogbookNode
from randovania.games.game import RandovaniaGame
from randovania.layout.layout_description import LayoutDescription

NON_MAJOR_PROGRESSION = [
    "Missile Expansion",
    "Energy Tank",
    "Energy Transfer Module",
    "Artifact",
    "Power Bomb Expansion",
    "Ice Spreader",
    "Wavebuster",
    "Gravity Suit",
    "Flamethrower",
    "X-Ray Visor",
    "Grapple Beam",
    "Thermal Visor",
    "Phazon Suit",
]


def is_non_major_progression(x: str):
    x = x.lower()
    for item in NON_MAJOR_PROGRESSION:
        if x == item.lower():
            return True
    return False


def iterate_with_log(x):
    try:
        import tqdm
        return tqdm.tqdm(x)
    except ImportError:
        print("WARNING: tqdm not found. Use `pip install tqdm` to have progress feedback.")
        return x


def read_json(path: Path) -> dict:
    with path.open() as x:
        return json.load(x)


_KEY_MATCH = re.compile(r"Key (\d+)")
_ARTIFACT_MATCH = re.compile(r"Artifact of (\w+)")
_PLAYER_MATCH = re.compile(r" for Player \d+")


def _filter_item_name(name: str) -> str:
    return _PLAYER_MATCH.sub("", _ARTIFACT_MATCH.sub("Artifact", _KEY_MATCH.sub("Key", name)))


def accumulate_results(game_modifications: dict,
                       items: Dict[str, Dict[str, int]],
                       locations: Dict[str, Dict[str, int]],
                       item_hints: Dict[str, Dict[str, int]],
                       location_hints: Dict[str, Dict[str, int]],

                       index_to_location: Dict[int, Tuple[str, str]],
                       logbook_to_name: Dict[str, str],
                       major_progression_items_only: bool,
                       ):
    for world_name, world_data in game_modifications["locations"].items():
        for area_name, item_name in world_data.items():
            area_name = f"{world_name}/{area_name}"
            item_name = _filter_item_name(item_name)
            if major_progression_items_only and is_non_major_progression(item_name):
                continue
            items[item_name][area_name] += 1
            locations[area_name][item_name] += 1

    for logbook_asset, hint_data in game_modifications["hints"].items():
        if hint_data["hint_type"] != "location":
            continue

        logbook_asset = logbook_to_name[logbook_asset]

        if hint_data["target"] == -1:
            item_name = "Nothing"
        else:
            area_name, location_name = index_to_location[hint_data["target"]]
            item_name = game_modifications["locations"][area_name][location_name]

        item_name = _filter_item_name(item_name)
        location_hints[logbook_asset][item_name] += 1
        item_hints[item_name][logbook_asset] += 1


def calculate_pickup_count(items: Dict[str, Dict[str, int]]) -> Dict[str, int]:
    return {
        name: sum(data.values())
        for name, data in items.items()
    }


def sort_by_contents(data: dict) -> dict:
    return {
        item: {
            location: count
            for location, count in sorted(data[item].items(), key=lambda t: t[1], reverse=True)
        }
        for item in sorted(data.keys())
    }


def calculate_stddev(pickup_count: Dict[str, int], item_counts: Dict[str, float]) -> Optional[float]:
    balanced_freq = {
        item: count / pickup_count[item]
        for item, count in item_counts.items()
        if item in pickup_count
    }
    try:
        return stdev(balanced_freq.values())
    except statistics.StatisticsError:
        return None


def first_key(d: dict):
    for key in d:
        return key


def get_items_order(all_items: Iterable[str], item_order: List[str], major_progression_items_only: bool) -> Tuple[
    Dict[str, int], Set[str], Set[str], Set[str]]:
    locations = set()
    no_key = set()
    progression_items = set()
    order = {}

    for i, entry in enumerate(item_order):
        splitter = " as "
        if splitter not in entry:
            splitter = " at "
        item, location = entry.split(splitter, 1)
        if major_progression_items_only and is_non_major_progression(item):
            continue
        order[item] = i
        location = location.split(" with ", 1)[0]
        locations.add(location)
        if "Key" not in item and "Artifact" not in item:
            no_key.add(location)
        progression_items.add(item)

    for item in all_items:
        if item not in order:
            order[item] = len(item_order)

    return order, locations, no_key, progression_items


def create_report(seeds_dir: str, output_file: str, csv_dir: Optional[str], use_percentage: bool,
                  major_progression_items_only: bool):
    def item_creator():
        return collections.defaultdict(int)

    items = collections.defaultdict(item_creator)
    locations = collections.defaultdict(item_creator)
    item_hints = collections.defaultdict(item_creator)
    location_hints = collections.defaultdict(item_creator)
    item_order = collections.defaultdict(list)

    game_description = game_description_for(RandovaniaGame.METROID_PRIME_ECHOES)
    world_list = game_description.world_list
    index_to_location = {
        node.pickup_index.index: (world_list.world_name_from_node(node, distinguish_dark_aether=True),
                                  world_list.node_name(node))
        for node in game_description.world_list.all_nodes
        if isinstance(node, PickupNode)
    }
    progression_count_for_location = collections.defaultdict(int)
    progression_no_key_count_for_location = collections.defaultdict(int)

    logbook_to_name = {
        str(node.string_asset_id): game_description.world_list.node_name(node)
        for node in game_description.world_list.all_nodes
        if isinstance(node, LogbookNode)
    }

    seed_count = 0
    pickup_count = None
    progression_items = None

    seed_files = list(Path(seeds_dir).glob(f"**/*.{LayoutDescription.file_extension()}"))
    seed_files.extend(Path(seeds_dir).glob("**/*.json"))
    for seed in iterate_with_log(seed_files):
        try:
            seed_data = read_json(seed)
        except json.JSONDecodeError:
            continue

        for game_modifications in seed_data["game_modifications"]:
            accumulate_results(game_modifications,
                               items, locations,
                               item_hints, location_hints,
                               index_to_location,
                               logbook_to_name,
                               major_progression_items_only)
        if seed_count == 0:
            pickup_count = calculate_pickup_count(items)

        item_orders, locations_with_progression, no_key_progression, _progression_items = get_items_order(
            list(items.keys()),
            seed_data["item_order"], major_progression_items_only)
        for item, order in item_orders.items():
            item_order[item].append(order)

        for location in locations_with_progression:
            progression_count_for_location[location] += 1

        for location in no_key_progression:
            progression_no_key_count_for_location[location] += 1

        if progression_items is None:
            progression_items = _progression_items

        seed_count += 1

    if pickup_count is None:
        raise Exception("No seeds found")

    stddev_by_location = {
        location: calculate_stddev(pickup_count, locations[location])
        for location in locations.keys()
    }

    regions = dict()
    region_totals = dict()

    total_progression_item_count = 0
    for location in locations:
        region = location.split("/")[0]

        if region not in regions.keys():
            regions[region] = 0

        if region not in region_totals.keys():
            region_totals[region] = 0
        region_totals[region] += 1

        count = 0
        for item in locations[location]:
            if (item in progression_items) and ("artifact" not in item.lower()) and ("key" not in item.lower()):
                count = count + locations[location][item]
        total_progression_item_count += count
        regions[region] += count

    # probability that any given location in this region contains progression
    regions_weighted = dict()
    for region in regions:
        regions_weighted[region] = (regions[region] / seed_count) / region_totals[region]

    items = sort_by_contents(items)
    locations = sort_by_contents(locations)
    item_hints = sort_by_contents(item_hints)
    location_hints = sort_by_contents(location_hints)

    location_progression_count = {
        location: value
        for location, value in sorted(progression_count_for_location.items(), key=lambda t: t[1], reverse=True)
    }

    location_progression_no_key_count = {
        location: value
        for location, value in sorted(progression_no_key_count_for_location.items(),
                                      key=lambda t: t[1], reverse=True)
    }

    for location in locations:
        if location not in location_progression_no_key_count.keys():
            location_progression_no_key_count[location] = 0

        if location not in location_progression_count.keys():
            location_progression_count[location] = 0

        for item in items:
            if location not in items[item].keys():
                items[item][location] = 0

    if use_percentage:
        for item in items:
            for room in items[item]:
                items[item][room] = items[item][room] / seed_count
        for location in locations:
            for item in locations[location]:
                locations[location][item] = locations[location][item] / seed_count
        for location in location_progression_count:
            location_progression_count[location] = location_progression_count[location] / seed_count
        for location in location_progression_no_key_count:
            location_progression_no_key_count[location] = location_progression_no_key_count[location] / seed_count
        for region in regions:
            regions[region] = regions[region] / total_progression_item_count

    final_results = {
        "seed_count": seed_count,
        "regions": regions,
        "regions_weighted": regions_weighted,
        "stddev_by_location": {
            location: stddev
            for location, stddev in sorted(stddev_by_location.items(), key=lambda t: t[1] or math.inf, reverse=True)
        },
        "items": items,
        "locations": locations,
        "item_hints": item_hints,
        "location_hints": location_hints,
        "location_progression_count": location_progression_count,
        "location_progression_no_key_count": location_progression_no_key_count,
        "item_order": {
            "average": {
                name: statistics.mean(orders)
                for name, orders in item_order.items()
            },
            "median": {
                name: int(statistics.median(orders))
                for name, orders in item_order.items()
            },
            "stdev": {
                name: statistics.stdev(orders)
                for name, orders in item_order.items()
            },
        }
    }

    # add non-item insights to csv file
    for location in locations:
        locations[location]["Progression Probability"] = location_progression_no_key_count[location]

    if csv_dir is not None:
        os.makedirs(csv_dir, exist_ok=True)
        for field in "items", "locations", "item_hints", "location_hints", "regions":
            data = final_results[field]

            if field == "regions":
                data = {
                    "Share of Progression": final_results["regions"],
                    "Locations with Progression": final_results["regions_weighted"],
                }

            possible_columns = set()
            for potential_values in data.values():
                possible_columns |= set(potential_values.keys())

            possible_columns = list(sorted(possible_columns))
            possible_columns.insert(0, "row_name")

            with open(os.path.join(csv_dir, field + ".csv"), "w", newline='') as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=possible_columns)
                writer.writeheader()

                for column, row_data in data.items():
                    row_data = copy.copy(row_data)
                    row_data["row_name"] = column
                    writer.writerow(row_data)

    with open(output_file, "w") as output:
        json.dump(final_results, output, indent=4, separators=(',', ': '))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--csv-dir")
    parser.add_argument("seeds_dir")
    parser.add_argument("output_file")
    parser.add_argument('--use-percentage', action='store_true')
    parser.add_argument('--major-progression-only', action='store_true')
    args = parser.parse_args()
    create_report(args.seeds_dir, args.output_file,
                  args.csv_dir, args.use_percentage,
                  args.major_progression_only)


if __name__ == "__main__":
    main()



import pytest

from randovania.layout.layout_description import LayoutDescription
from randovania.resolver import resolver, debug


@pytest.mark.skip_resolver_tests
@pytest.mark.parametrize("seed_name", ["seed_a.rdvgame", "corruption_seed_a.rdvgame"])
@pytest.mark.asyncio
async def test_resolver_with_log_file(test_files_dir, seed_name: str):
    # Setup
    debug.set_level(2)

    description = LayoutDescription.from_file(test_files_dir.joinpath("log_files", seed_name))
    configuration = description.permalink.presets[0].configuration
    patches = description.all_patches[0]

    # Run
    final_state_by_resolve = await resolver.resolve(configuration=configuration,
                                                    patches=patches)

    # Assert
    assert final_state_by_resolve is not None



from dataclasses import dataclass
from typing import Dict, Optional, Tuple

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame


@dataclass(frozen=True)
class MajorItem:
    game: RandovaniaGame
    name: str
    item_category: ItemCategory
    broad_category: ItemCategory
    model_name: str
    progression: Tuple[int, ...]
    ammo_index: Tuple[int, ...] = tuple()
    unlocks_ammo: bool = False
    required: bool = False
    original_index: Optional[PickupIndex] = None
    probability_offset: int = 0
    probability_multiplier: float = 1
    warning: Optional[str] = None

    def __post_init__(self):
        if not self.progression and not self.ammo_index:
            raise ValueError(f"Item {self.name} has no progression nor ammo.")

    @classmethod
    def from_json(cls, name: str, value: dict, game: RandovaniaGame,
                  item_categories: Dict[str, ItemCategory]) -> "MajorItem":
        return cls(
            game=game,
            name=name,
            item_category=item_categories[value["item_category"]],
            broad_category=item_categories[value["broad_category"]],
            model_name=value["model_name"],
            progression=tuple(value["progression"]),
            ammo_index=tuple(value.get("ammo", [])),
            unlocks_ammo=value.get("unlocks_ammo", False),
            required=value.get("required", False),
            original_index=PickupIndex(value["original_index"]) if "original_index" in value else None,
            probability_offset=value["probability_offset"],
            probability_multiplier=value["probability_multiplier"],
            warning=value.get("warning"),
        )

    @property
    def as_json(self) -> dict:
        result = {
            "item_category": self.item_category.name,
            "broad_category": self.broad_category.name,
            "model_name": self.model_name,
            "progression": list(self.progression),
            "ammo": list(self.ammo_index),
            "unlocks_ammo": self.unlocks_ammo,
            "required": self.required,
            "probability_offset": self.probability_offset,
            "probability_multiplier": self.probability_multiplier,
        }
        if self.original_index is not None:
            result["original_index"] = self.original_index.index
        if self.warning is not None:
            result["warning"] = self.warning
        return result



import copy
import io
import json
import math
import uuid
from pathlib import Path
from typing import Optional, Dict

import aiofiles
import slugify

from randovania.game_description.world.area_location import AreaLocation
from randovania.games.game import RandovaniaGame
from randovania.layout.preset import Preset

CURRENT_PRESET_VERSION = 14


class InvalidPreset(Exception):
    def __init__(self, original_exception: Exception):
        self.original_exception = original_exception


def _migrate_v1(preset: dict) -> dict:
    layout_configuration = preset["layout_configuration"]
    layout_configuration["beam_configuration"] = {
        "power": {
            "item_index": 0,
            "ammo_a": -1,
            "ammo_b": -1,
            "uncharged_cost": 0,
            "charged_cost": 0,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 0
        },
        "dark": {
            "item_index": 1,
            "ammo_a": 45,
            "ammo_b": -1,
            "uncharged_cost": 1,
            "charged_cost": 5,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 30
        },
        "light": {
            "item_index": 2,
            "ammo_a": 46,
            "ammo_b": -1,
            "uncharged_cost": 1,
            "charged_cost": 5,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 30
        },
        "annihilator": {
            "item_index": 3,
            "ammo_a": 46,
            "ammo_b": 45,
            "uncharged_cost": 1,
            "charged_cost": 5,
            "combo_missile_cost": 5,
            "combo_ammo_cost": 30
        }
    }
    layout_configuration["skip_final_bosses"] = False
    layout_configuration["energy_per_tank"] = 100
    return preset


def _migrate_v2(preset: dict) -> dict:
    level_renaming = {
        "trivial": "beginner",
        "easy": "intermediate",
        "normal": "advanced",
        "hard": "expert",
        "minimal-restrictions": "minimal-logic",
    }
    trick_level = preset["layout_configuration"]["trick_level"]
    trick_level["global_level"] = level_renaming.get(trick_level["global_level"], trick_level["global_level"])
    for specific, value in trick_level["specific_levels"].items():
        trick_level["specific_levels"][specific] = level_renaming.get(value, value)

    return preset


def _migrate_v3(preset: dict) -> dict:
    preset["layout_configuration"]["safe_zone"] = {
        "fully_heal": True,
        "prevents_dark_aether": True,
        "heal_per_second": 1.0,
    }
    return preset


def _migrate_v4(preset: dict) -> dict:
    trick_name_mapping = {
        0: "Dash",
        1: "BombJump",
        2: "SlopeJump",
        3: "Movement",
        4: "BSJ",
        5: "RollJump",
        6: "UnderwaterDash",
        7: "AirUnderwater",
        8: "OoB",
        10: "SAnoSJ",
        11: "WallBoost",
        12: "EnemyHop",
        13: "Combat",
        15: "InstantMorph",
        26: "InvisibleObjects",
        27: "StandableTerrain",
        28: "TerminalFall",
        29: "BoostJump",
        30: "EDash",
        31: "BomblessSlot",
        32: "ScanPost",
        33: "ScrewAttackTunnels",
        34: "Knowledge",
        35: "SeekerlessLocks",
    }

    preset["layout_configuration"]["game"] = "prime2"

    trick_level = preset["layout_configuration"]["trick_level"]
    global_level = trick_level.pop("global_level")

    trick_level["minimal_logic"] = global_level == "minimal-logic"
    specific_levels = {
        trick_name_mapping[int(index)]: level
        for index, level in trick_level["specific_levels"].items()
    }
    if global_level == "minimal-logic":
        specific_levels = {}
    elif global_level == "beginner":
        for trick in ["BombJump", "Knowledge", "Movement", "ScanPost", "SAnoSJ"]:
            specific_levels[trick] = "beginner"
    else:
        for trick in trick_name_mapping.values():
            if trick not in specific_levels:
                specific_levels[trick] = global_level

    trick_level["specific_levels"] = specific_levels

    return preset


def _migrate_v5(preset: dict) -> dict:
    excluded_item = {
        "include_copy_in_original_location": False,
        "num_shuffled_pickups": 0,
        "num_included_in_starting_items": 0,
        "included_ammo": [],
        "allowed_as_random_starting_item": True,
    }
    included_item = {
        **excluded_item,
        "num_included_in_starting_items": 1
    }
    shuffled_item = {
        **excluded_item,
        "num_shuffled_pickups": 1,
    }

    default_items_state = {
        "Progressive Suit": {**excluded_item, "num_shuffled_pickups": 2},
        "Dark Beam": {**shuffled_item, "included_ammo": [50]},
        "Light Beam": {**shuffled_item, "included_ammo": [50]},
        "Annihilator Beam": {**shuffled_item, "included_ammo": [0, 0]},
        "Power Bomb": {**shuffled_item, "included_ammo": [2]},
        "Progressive Grapple": {**excluded_item, "num_shuffled_pickups": 2},
        "Missile Launcher": {**shuffled_item, "included_ammo": [5]},
        "Seeker Launcher": {**shuffled_item, "included_ammo": [5]},
        "Energy Tank": {**excluded_item, "num_shuffled_pickups": 14},
    }

    for item in ["Combat Visor", "Scan Visor", "Varia Suit", "Power Beam", "Charge Beam", "Morph Ball"]:
        default_items_state[item] = included_item
    for item in ["Dark Visor", "Echo Visor", "Morph Ball Bomb", "Boost Ball", "Spider Ball", "Space Jump Boots",
                 "Gravity Boost", "Super Missile", "Sunburst", "Darkburst", "Sonic Boom", "Violet Translator",
                 "Amber Translator", "Emerald Translator", "Cobalt Translator"]:
        default_items_state[item] = shuffled_item

    major_items = preset["layout_configuration"]["major_items_configuration"]["items_state"]
    for item in default_items_state.keys():
        if item not in major_items:
            major_items[item] = default_items_state[item]

    preset["layout_configuration"]["major_items_configuration"].pop("progressive_suit")
    preset["layout_configuration"]["major_items_configuration"].pop("progressive_grapple")
    preset["layout_configuration"].pop("split_beam_ammo")

    specific_levels: Dict[str, str] = preset["layout_configuration"]["trick_level"]["specific_levels"]
    tricks_to_remove = [trick_name for trick_name, level in specific_levels.items() if level == "no-tricks"]
    for trick in tricks_to_remove:
        specific_levels.pop(trick)

    preset["game"] = preset["layout_configuration"].pop("game")
    preset["configuration"] = preset.pop("layout_configuration")
    preset["configuration"].update(preset.pop("patcher_configuration"))
    preset["configuration"]["varia_suit_damage"] = max(preset["configuration"]["varia_suit_damage"], 0.1)

    return preset


def _migrate_v6(preset: dict) -> dict:
    preset["configuration"]["dangerous_energy_tank"] = False
    return preset


def _migrate_v7(preset: dict) -> dict:
    default_items = {}
    if preset["game"] == "prime2":
        default_items["visor"] = "Combat Visor"
        default_items["beam"] = "Power Beam"

    preset["configuration"]["major_items_configuration"]["default_items"] = default_items
    return preset


def _migrate_v8(preset: dict) -> dict:
    from randovania.game_description import default_database
    game = default_database.game_description_for(RandovaniaGame(preset["game"]))

    def _name_to_location(name: str):
        world_name, area_name = name.split("/", 1)
        world = game.world_list.world_with_name(world_name)
        area = world.area_by_name(area_name)
        return AreaLocation(world.world_asset_id, area.area_asset_id)

    preset["configuration"]["multi_pickup_placement"] = False

    if "energy_per_tank" in preset["configuration"]:
        preset["configuration"]["energy_per_tank"] = int(preset["configuration"]["energy_per_tank"])

    preset["configuration"]["starting_location"] = [
        _name_to_location(location).as_json
        for location in preset["configuration"]["starting_location"]
    ]

    excluded_teleporters = []
    if preset["game"] == "prime2":
        excluded_teleporters = [
            {
                "world_asset_id": 464164546,
                "area_asset_id": 3136899603,
                "instance_id": 204865660
            },
            {
                "world_asset_id": 2252328306,
                "area_asset_id": 2068511343,
                "instance_id": 589949
            },
            {
                "world_asset_id": 464164546,
                "area_asset_id": 1564082177,
                "instance_id": 4260106
            },
            {
                "world_asset_id": 1006255871,
                "area_asset_id": 2278776548,
                "instance_id": 136970379
            },
        ]

    preset["configuration"]["elevators"] = {
        "mode": preset["configuration"]["elevators"],
        "excluded_teleporters": excluded_teleporters,
        "excluded_targets": [],
        "skip_final_bosses": preset["configuration"].pop("skip_final_bosses", False),
        "allow_unvisited_room_names": True,
    }
    if preset["configuration"]["major_items_configuration"]["default_items"].get("visor") == "Dark Visor":
        preset["configuration"]["major_items_configuration"]["default_items"]["visor"] = "Combat Visor"
        preset["configuration"]["major_items_configuration"]["items_state"]["Scan Visor"] = {
            "num_included_in_starting_items": 1
        }

    return preset


def _migrate_v9(preset: dict) -> dict:
    if preset.get("uuid") is None:
        preset["uuid"] = str(uuid.uuid4())

    _name_to_uuid = {
        "Corruption Preset": "5682c9ef-d447-4327-b473-ba1216d83439",
        "Darkszero's Deluxe": "ea1eced4-53b8-4bb3-a08f-27ac1afe6aab",
        "Fewest Changes": "bba48268-0710-4ac5-baae-dcd5fcd31d80",
        "Prime Preset": "e36f52b3-ccd9-4dd7-a18f-b57b25f6b079",
        "Starter Preset": "fcbe4e3f-b1fa-41cc-83cb-c86d84c10f0f",
    }

    base_preset_name = preset.pop("base_preset_name")
    preset["base_preset_uuid"] = _name_to_uuid.get(base_preset_name, str(uuid.uuid4()))

    if preset["game"] != "prime2":
        preset["configuration"].pop("dangerous_energy_tank")

    return preset


def _migrate_v10(preset: dict) -> dict:
    if preset["game"] == "prime1":
        major = preset["configuration"].pop("qol_major_cutscenes")
        minor = preset["configuration"].pop("qol_minor_cutscenes")
        if major:
            cutscenes = "major"
        elif minor:
            cutscenes = "minor"
        else:
            cutscenes = "original"
        preset["configuration"]["qol_cutscenes"] = cutscenes

    elif preset["game"] == "prime2":
        preset["configuration"]["allow_jumping_on_dark_water"] = False
        fields = [
            "allow_vanilla_dark_beam",
            "allow_vanilla_light_beam",
            "allow_vanilla_seeker_launcher",
            "allow_vanilla_echo_visor",
            "allow_vanilla_dark_visor",
            "allow_vanilla_screw_attack",
            "allow_vanilla_gravity_boost",
            "allow_vanilla_boost_ball",
            "allow_vanilla_spider_ball",
        ]
        for f in fields:
            preset["configuration"][f] = True

    return preset


def _migrate_v11(preset: dict) -> dict:
    if preset["game"] == "prime1":
        preset["configuration"]["warp_to_start"] = False

    return preset


def _migrate_v12(preset: dict) -> dict:
    preset["configuration"]["logical_resource_action"] = "randomly"
    if preset["game"] == "prime1":
        preset["configuration"]["artifact_target"] = preset["configuration"].pop("artifacts")
        preset["configuration"]["artifact_minimum_progression"] = 0
        preset["configuration"]["qol_pickup_scans"] = False

    return preset


def _migrate_v13(preset: dict) -> dict:
    for config in preset["configuration"]["major_items_configuration"]["items_state"].values():
        config.pop("allowed_as_random_starting_item", None)

    maximum_ammo = preset["configuration"]["ammo_configuration"].pop("maximum_ammo")
    ammo_ids = {
        "prime1": {
            "Missile Expansion": ["4"],
            "Power Bomb Expansion": ["7"],
        },
        "prime2": {
            "Missile Expansion": ["44"],
            "Power Bomb Expansion": ["43"],
            "Dark Ammo Expansion": ["45"],
            "Light Ammo Expansion": ["46"],
            "Beam Ammo Expansion": ["45", "46"],
        },
        "prime3": {
            "Missile Expansion": ["4"],
            "Ship Missile Expansion": ["45"],
        },
        "super_metroid": {
            "Missile Expansion": ["5"],
            "Super Missile Expansion": ["6"],
            "Power Bomb Expansion": ["7"],
        },
    }[preset["game"]]
    main_items = {
        "prime1": {
            "Missile Launcher": ["4"],
            "Power Bomb": ["7"],
        },
        "prime2": {
            "Dark Beam": ["45"],
            "Light Beam": ["46"],
            "Annihilator Beam": ["45", "46"],
            "Missile Launcher": ["44"],
            "Seeker Launcher": ["44"],
            "Power Bomb": ["43"],
        },
        "prime3": {
            "Missile Launcher": ["4"],
            "Ship Missile": ["45"],
        },
        "super_metroid": {},
    }

    for item, ids in main_items[preset["game"]].items():
        item_state = preset["configuration"]["major_items_configuration"]["items_state"][item]
        count = item_state.get("num_shuffled_pickups", 0) + item_state.get("num_included_in_starting_items", 0)
        if item_state.get("include_copy_in_original_location", False):
            count += 1

        for ammo_id, ammo in zip(ids, item_state["included_ammo"]):
            maximum_ammo[ammo_id] -= ammo * count

    for name, config in preset["configuration"]["ammo_configuration"]["items_state"].items():
        config["ammo_count"] = [
            math.ceil(max(maximum_ammo[ammo_id], 0) / max(config["pickup_count"], 1))
            for ammo_id in ammo_ids[name]
        ]
        config.pop("variance")

    return preset


_MIGRATIONS = {
    1: _migrate_v1,
    2: _migrate_v2,
    3: _migrate_v3,
    4: _migrate_v4,
    5: _migrate_v5,
    6: _migrate_v6,
    7: _migrate_v7,
    8: _migrate_v8,
    9: _migrate_v9,
    10: _migrate_v10,
    11: _migrate_v11,
    12: _migrate_v12,
    13: _migrate_v13,
}


def _apply_migration(preset: dict, version: int) -> dict:
    while version < CURRENT_PRESET_VERSION:
        preset = _MIGRATIONS[version](preset)
        version += 1
    return preset


def convert_to_current_version(preset: dict) -> dict:
    schema_version = preset["schema_version"]
    if schema_version > CURRENT_PRESET_VERSION:
        raise ValueError(f"Unknown version: {schema_version}")

    if schema_version < CURRENT_PRESET_VERSION:
        return _apply_migration(preset, schema_version)
    else:
        return preset


class VersionedPreset:
    data: dict
    exception: Optional[InvalidPreset] = None
    _preset: Optional[Preset] = None

    def __init__(self, data):
        self.data = data

    @classmethod
    def file_extension(cls) -> str:
        return "rdvpreset"

    @property
    def slug_name(self) -> str:
        return slugify.slugify(self.name)

    @property
    def name(self) -> str:
        if self._preset is not None:
            return self._preset.name
        else:
            return self.data["name"]

    @property
    def base_preset_uuid(self) -> Optional[uuid.UUID]:
        if self._preset is not None:
            return self._preset.base_preset_uuid
        elif self.data["base_preset_uuid"] is not None:
            return uuid.UUID(self.data["base_preset_uuid"])

    @property
    def game(self) -> RandovaniaGame:
        if self._preset is not None:
            return self._preset.configuration.game

        if self.data["schema_version"] < 6:
            return RandovaniaGame.METROID_PRIME_ECHOES

        return RandovaniaGame(self.data["game"])

    @property
    def uuid(self) -> uuid.UUID:
        if self._preset is not None:
            return self._preset.uuid
        else:
            return uuid.UUID(self.data["uuid"])

    def __eq__(self, other):
        if isinstance(other, VersionedPreset):
            return self.get_preset() == other.get_preset()
        return False

    def is_for_known_game(self):
        try:
            # self.game is never None, but it might raise ValueError in case the preset is for an unknown game
            return self.game is not None
        except ValueError:
            return False

    @property
    def _converted(self):
        return self._preset is not None or self.exception is not None

    def ensure_converted(self):
        if not self._converted:
            try:
                self._preset = Preset.from_json_dict(convert_to_current_version(copy.deepcopy(self.data)))
            except (ValueError, KeyError, TypeError) as e:
                self.exception = InvalidPreset(e)
                raise self.exception from e

    def get_preset(self) -> Preset:
        self.ensure_converted()
        if self.exception:
            raise self.exception
        else:
            return self._preset

    @classmethod
    async def from_file(cls, path: Path) -> "VersionedPreset":
        async with aiofiles.open(path) as f:
            return VersionedPreset(json.loads(await f.read()))

    @classmethod
    def from_file_sync(cls, path: Path) -> "VersionedPreset":
        with path.open() as f:
            return VersionedPreset(json.load(f))

    @classmethod
    def with_preset(cls, preset: Preset) -> "VersionedPreset":
        result = VersionedPreset(None)
        result._preset = preset
        return result

    def save_to_file(self, path: Path):
        path.parent.mkdir(exist_ok=True, parents=True)
        with path.open("w") as preset_file:
            json.dump(self.as_json, preset_file, indent=4)

    def save_to_io(self, data: io.BytesIO):
        data.write(
            json.dumps(self.as_json, indent=4).encode("utf-8")
        )

    @property
    def as_json(self) -> dict:
        if self._preset is not None:
            preset_json = {
                "schema_version": CURRENT_PRESET_VERSION,
            }
            preset_json.update(self._preset.as_json)
            return preset_json
        else:
            assert self.data is not None
            return self.data



import base64
from pathlib import Path

import pytest
import socketio.exceptions
from mock import MagicMock, AsyncMock, call

from randovania.game_description.resources.pickup_entry import PickupEntry, PickupModel
from randovania.games.game import RandovaniaGame
from randovania.network_client.game_session import GameSessionPickups
from randovania.network_client.network_client import NetworkClient, ConnectionState, _decode_pickup
from randovania.network_common import connection_headers
from randovania.network_common.admin_actions import SessionAdminGlobalAction, SessionAdminUserAction
from randovania.network_common.error import InvalidSession, RequestTimeout, ServerError


@pytest.fixture(name="client")
def _client(tmpdir):
    return NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000"})


@pytest.mark.asyncio
async def test_on_connect_no_restore(tmpdir):
    client = NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000"})

    # Run
    await client.on_connect()

    # Assert
    assert client.connection_state == ConnectionState.ConnectedNotLogged


@pytest.mark.parametrize("valid_session", [False, True])
@pytest.mark.asyncio
async def test_on_connect_restore(tmpdir, valid_session: bool):
    client = NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000"})
    session_data_path = Path(tmpdir) / "9iwAGnskOkqzo_NZ" / "session_persistence.bin"
    session_data_path.parent.mkdir(parents=True)
    session_data_path.write_bytes(b"foo")

    if valid_session:
        call_result = {"result": {
            "user": {
                "id": 1234,
                "name": "You",
            },
            "encoded_session_b85": b'Ze@30VtI6Ba{'
        }}
    else:
        call_result = InvalidSession().as_json

    client.sio = MagicMock()
    client.sio.call = AsyncMock(return_value=call_result)

    # Run
    await client.on_connect()

    # Assert
    client.sio.call.assert_awaited_once_with("restore_user_session", (b"foo", None), namespace=None, timeout=30)

    if valid_session:
        assert client.connection_state == ConnectionState.Connected
        assert session_data_path.read_bytes() == b"new_bytes"
    else:
        assert client.connection_state == ConnectionState.ConnectedNotLogged
        assert not session_data_path.is_file()


@pytest.mark.asyncio
async def test_on_connect_restore_timeout(client: NetworkClient):
    # Setup
    client._restore_session = AsyncMock(side_effect=ServerError())
    client._connection_state = ConnectionState.Connecting
    client.disconnect_from_server = AsyncMock()

    # Run
    await client.on_connect()

    # Assert
    assert client.connection_state == ConnectionState.Disconnected
    client.disconnect_from_server.assert_awaited_once_with()


@pytest.mark.asyncio
async def test_connect_to_server(tmpdir):
    # Setup
    client = NetworkClient(Path(tmpdir), {"server_address": "http://localhost:5000",
                                          "socketio_path": "/path"})

    async def connect(*args, **kwargs):
        client._waiting_for_on_connect.set_result(True)

    client.sio = MagicMock()
    client.sio.connect = AsyncMock(side_effect=connect)
    client.sio.connected = False

    # Run
    await client.connect_to_server()

    # Assert
    assert client.connection_state == ConnectionState.Connecting
    client.sio.connect.assert_awaited_once_with("http://localhost:5000",
                                                socketio_path="/path",
                                                transports=["websocket"],
                                                headers=connection_headers())


@pytest.mark.asyncio
async def test_session_admin_global(client):
    client._emit_with_result = AsyncMock()
    client._current_game_session_meta = MagicMock()
    client._current_game_session_meta.id = 1234

    # Run
    result = await client.session_admin_global(SessionAdminGlobalAction.CHANGE_ROW, 5)

    # Assert
    assert result == client._emit_with_result.return_value
    client._emit_with_result.assert_awaited_once_with("game_session_admin_session", (1234, "change_row", 5))


@pytest.mark.parametrize("permanent", [False, True])
@pytest.mark.asyncio
async def test_leave_game_session(client: NetworkClient, permanent: bool):
    client._emit_with_result = AsyncMock()
    client._current_game_session_meta = MagicMock()
    client._current_game_session_meta.id = 1234
    client._current_user = MagicMock()
    client._current_user.id = 5678

    # Run
    await client.leave_game_session(permanent)

    # Assert
    calls = [call("disconnect_game_session", 1234)]
    if permanent:
        calls.insert(0, call("game_session_admin_player", (1234, 5678, SessionAdminUserAction.KICK.value, None)))

    client._emit_with_result.assert_has_awaits(calls)

    assert client._current_game_session_meta is None


@pytest.mark.asyncio
async def test_emit_with_result_timeout(client: NetworkClient):
    # Setup
    client._connection_state = ConnectionState.Connected
    client.sio = AsyncMock()
    client.sio.call.side_effect = socketio.exceptions.TimeoutError()

    # Run
    with pytest.raises(RequestTimeout, match="Timeout after "):
        await client._emit_with_result("test_event")


def test_update_timeout_with_increase(client: NetworkClient):
    # Run
    client._update_timeout_with(5.0, False)

    # Assert
    assert client._current_timeout == 40


def test_update_timeout_with_dont_decrease_below_minimum(client: NetworkClient):
    # Setup
    client._current_timeout = 30

    # Run
    client._update_timeout_with(5.0, True)

    # Assert
    assert client._current_timeout == 30


def test_update_timeout_with_decrease_on_success(client: NetworkClient):
    # Setup
    client._current_timeout = 50

    # Run
    client._update_timeout_with(5.0, True)

    # Assert
    assert client._current_timeout == 40


@pytest.mark.asyncio
async def test_refresh_received_pickups(client: NetworkClient, corruption_game_description, mocker):
    db = corruption_game_description.resource_database

    data = {
        "game": RandovaniaGame.METROID_PRIME_CORRUPTION.value,
        "pickups": [
            {"provider_name": "Message A", "pickup": 'VtI6Bb3p'},
            {"provider_name": "Message B", "pickup": 'VtI6Bb3y'},
            {"provider_name": "Message C", "pickup": 'VtI6Bb3*'},
        ]
    }

    pickups = [MagicMock(), MagicMock(), MagicMock()]
    mock_decode = mocker.patch("randovania.network_client.network_client._decode_pickup", side_effect=pickups)

    # Run
    await client._on_game_session_pickups_update_raw(data)

    # Assert
    assert client._current_game_session_pickups == GameSessionPickups(
        game=RandovaniaGame.METROID_PRIME_CORRUPTION,
        pickups=(
            ("Message A", pickups[0]),
            ("Message B", pickups[1]),
            ("Message C", pickups[2]),
        )
    )
    mock_decode.assert_has_calls([call('VtI6Bb3p', db), call('VtI6Bb3y', db), call('VtI6Bb3*', db)])


def test_decode_pickup(client: NetworkClient, echoes_resource_database, generic_item_category):
    data = "h^WxYK%Bzb%FGCg+{)z6@~Gs(+~&;6<eJ{d>e$N-ot3$j$)%29RCZS8S0<J+1eZou(gZdaMpkzBCzit8K`;ZWbDM)J8%r}wD<K0*qbq3*ot3$j$)%29RCZS8S0<J+1eZou(gZdaMpkzBCzit8K`;ZWbDM)J8%r}wD<K0*qbq0uzy"
    expected_pickup = PickupEntry(
        name="The Name",
        model=PickupModel(
            game=RandovaniaGame.METROID_PRIME_ECHOES,
            name="EnergyTransferModule",
        ),
        item_category=generic_item_category,
        broad_category=generic_item_category, 
        progression=tuple(),
    )
    
    # Uncomment this to encode the data once again and get the new bytefield if it changed for some reason 
#    from randovania.server.game_session import _base64_encode_pickup
#    new_data = _base64_encode_pickup(expected_pickup, echoes_resource_database)
#    assert new_data == data

    # Run
    pickup = _decode_pickup(data, echoes_resource_database)

    # Assert
    assert pickup == expected_pickup



from unittest.mock import MagicMock, call

import randovania.generator.item_pool.ammo
import randovania.generator.item_pool.pickup_creator
from randovania.layout.base.ammo_configuration import AmmoConfiguration


def test_add_ammo(echoes_resource_database, mocker):
    # Setup
    mock_create_ammo_expansion: MagicMock = mocker.patch(
        "randovania.generator.item_pool.ammo.create_ammo_expansion", autospec=True)

    ammo1 = MagicMock()
    ammo2 = MagicMock()
    state1 = MagicMock()
    state2 = MagicMock()

    ammo_configuration = AmmoConfiguration({
        ammo1: state1,
        ammo2: state2,
    })

    # Run
    results = list(randovania.generator.item_pool.ammo.add_ammo(echoes_resource_database, ammo_configuration))

    # Assert
    assert results == [mock_create_ammo_expansion.return_value, mock_create_ammo_expansion.return_value]
    mock_create_ammo_expansion.assert_has_calls([
        call(ammo1, state1.ammo_count, state1.requires_major_item, echoes_resource_database),
        call(ammo2, state2.ammo_count, state2.requires_major_item, echoes_resource_database),
    ])



from pathlib import Path

from appdirs import AppDirs

local_dirs = AppDirs("Randovania", False, roaming=False)
roaming_dirs = AppDirs("Randovania", False, roaming=True)


def local_data_dir() -> Path:
    return Path(local_dirs.user_data_dir)


def roaming_data_dir() -> Path:
    return Path(roaming_dirs.user_config_dir)




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'game_session.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_GameSessionWindow(object):
    def setupUi(self, GameSessionWindow):
        if not GameSessionWindow.objectName():
            GameSessionWindow.setObjectName(u"GameSessionWindow")
        GameSessionWindow.resize(739, 422)
        GameSessionWindow.setDockNestingEnabled(True)
        self.action_add_player = QAction(GameSessionWindow)
        self.action_add_player.setObjectName(u"action_add_player")
        self.action_add_row = QAction(GameSessionWindow)
        self.action_add_row.setObjectName(u"action_add_row")
        self.rename_session_action = QAction(GameSessionWindow)
        self.rename_session_action.setObjectName(u"rename_session_action")
        self.change_password_action = QAction(GameSessionWindow)
        self.change_password_action.setObjectName(u"change_password_action")
        self.delete_session_action = QAction(GameSessionWindow)
        self.delete_session_action.setObjectName(u"delete_session_action")
        self.actionbar = QAction(GameSessionWindow)
        self.actionbar.setObjectName(u"actionbar")
        self.actionasdf = QAction(GameSessionWindow)
        self.actionasdf.setObjectName(u"actionasdf")
        self.central_widget = QWidget(GameSessionWindow)
        self.central_widget.setObjectName(u"central_widget")
        self.central_widget.setMaximumSize(QSize(16777215, 16777215))
        self.horizontalLayout = QHBoxLayout(self.central_widget)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setContentsMargins(11, 11, 11, 11)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.game_connection_tool = QToolButton(self.central_widget)
        self.game_connection_tool.setObjectName(u"game_connection_tool")
        self.game_connection_tool.setPopupMode(QToolButton.InstantPopup)

        self.horizontalLayout.addWidget(self.game_connection_tool)

        self.game_connection_label = QLabel(self.central_widget)
        self.game_connection_label.setObjectName(u"game_connection_label")

        self.horizontalLayout.addWidget(self.game_connection_label)

        self.connection_line_1 = QFrame(self.central_widget)
        self.connection_line_1.setObjectName(u"connection_line_1")
        self.connection_line_1.setFrameShape(QFrame.VLine)
        self.connection_line_1.setFrameShadow(QFrame.Sunken)

        self.horizontalLayout.addWidget(self.connection_line_1)

        self.server_connection_button = QPushButton(self.central_widget)
        self.server_connection_button.setObjectName(u"server_connection_button")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.server_connection_button.sizePolicy().hasHeightForWidth())
        self.server_connection_button.setSizePolicy(sizePolicy)
        self.server_connection_button.setMaximumSize(QSize(60, 16777215))

        self.horizontalLayout.addWidget(self.server_connection_button)

        self.server_connection_label = QLabel(self.central_widget)
        self.server_connection_label.setObjectName(u"server_connection_label")

        self.horizontalLayout.addWidget(self.server_connection_label)

        self.connection_line_2 = QFrame(self.central_widget)
        self.connection_line_2.setObjectName(u"connection_line_2")
        self.connection_line_2.setFrameShape(QFrame.VLine)
        self.connection_line_2.setFrameShadow(QFrame.Sunken)

        self.horizontalLayout.addWidget(self.connection_line_2)

        self.session_status_tool = QToolButton(self.central_widget)
        self.session_status_tool.setObjectName(u"session_status_tool")
        self.session_status_tool.setMaximumSize(QSize(80, 16777215))
        self.session_status_tool.setPopupMode(QToolButton.MenuButtonPopup)

        self.horizontalLayout.addWidget(self.session_status_tool)

        self.session_status_label = QLabel(self.central_widget)
        self.session_status_label.setObjectName(u"session_status_label")

        self.horizontalLayout.addWidget(self.session_status_label)

        self.connection_line_3 = QFrame(self.central_widget)
        self.connection_line_3.setObjectName(u"connection_line_3")
        self.connection_line_3.setFrameShape(QFrame.VLine)
        self.connection_line_3.setFrameShadow(QFrame.Sunken)

        self.horizontalLayout.addWidget(self.connection_line_3)

        self.advanced_options_tool = QToolButton(self.central_widget)
        self.advanced_options_tool.setObjectName(u"advanced_options_tool")
        self.advanced_options_tool.setPopupMode(QToolButton.InstantPopup)

        self.horizontalLayout.addWidget(self.advanced_options_tool)

        GameSessionWindow.setCentralWidget(self.central_widget)
        self.menu_bar = QMenuBar(GameSessionWindow)
        self.menu_bar.setObjectName(u"menu_bar")
        self.menu_bar.setGeometry(QRect(0, 0, 739, 22))
        GameSessionWindow.setMenuBar(self.menu_bar)
        self.players_dock = QDockWidget(GameSessionWindow)
        self.players_dock.setObjectName(u"players_dock")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.players_dock.sizePolicy().hasHeightForWidth())
        self.players_dock.setSizePolicy(sizePolicy1)
        self.players_dock.setFeatures(QDockWidget.DockWidgetFloatable|QDockWidget.DockWidgetMovable)
        self.players_dock_contents = QWidget()
        self.players_dock_contents.setObjectName(u"players_dock_contents")
        self.players_dock_layout = QVBoxLayout(self.players_dock_contents)
        self.players_dock_layout.setSpacing(0)
        self.players_dock_layout.setContentsMargins(11, 11, 11, 11)
        self.players_dock_layout.setObjectName(u"players_dock_layout")
        self.players_dock_layout.setContentsMargins(0, 0, 0, 0)
        self.players_box_scroll = QScrollArea(self.players_dock_contents)
        self.players_box_scroll.setObjectName(u"players_box_scroll")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.players_box_scroll.sizePolicy().hasHeightForWidth())
        self.players_box_scroll.setSizePolicy(sizePolicy2)
        self.players_box_scroll.setWidgetResizable(True)
        self.players_box = QWidget()
        self.players_box.setObjectName(u"players_box")
        self.players_box.setGeometry(QRect(0, 0, 233, 176))
        self.players_layout = QGridLayout(self.players_box)
        self.players_layout.setSpacing(6)
        self.players_layout.setContentsMargins(11, 11, 11, 11)
        self.players_layout.setObjectName(u"players_layout")
        self.presets_line = QFrame(self.players_box)
        self.presets_line.setObjectName(u"presets_line")
        self.presets_line.setFrameShape(QFrame.HLine)
        self.presets_line.setFrameShadow(QFrame.Sunken)

        self.players_layout.addWidget(self.presets_line, 1, 0, 1, 2)

        self.title_connection_state_label = QLabel(self.players_box)
        self.title_connection_state_label.setObjectName(u"title_connection_state_label")

        self.players_layout.addWidget(self.title_connection_state_label, 0, 4, 1, 1)

        self.players_vertical_line = QFrame(self.players_box)
        self.players_vertical_line.setObjectName(u"players_vertical_line")
        self.players_vertical_line.setFrameShape(QFrame.VLine)
        self.players_vertical_line.setFrameShadow(QFrame.Sunken)

        self.players_layout.addWidget(self.players_vertical_line, 0, 2, 1, 1)

        self.team_line = QFrame(self.players_box)
        self.team_line.setObjectName(u"team_line")
        self.team_line.setFrameShape(QFrame.HLine)
        self.team_line.setFrameShadow(QFrame.Sunken)

        self.players_layout.addWidget(self.team_line, 1, 3, 1, 2)

        self.new_row_button = QPushButton(self.players_box)
        self.new_row_button.setObjectName(u"new_row_button")

        self.players_layout.addWidget(self.new_row_button, 0, 0, 1, 2)

        self.title_player_name_label = QLabel(self.players_box)
        self.title_player_name_label.setObjectName(u"title_player_name_label")

        self.players_layout.addWidget(self.title_player_name_label, 0, 3, 1, 1)

        self.players_box_scroll.setWidget(self.players_box)

        self.players_dock_layout.addWidget(self.players_box_scroll)

        self.players_dock.setWidget(self.players_dock_contents)
        GameSessionWindow.addDockWidget(Qt.TopDockWidgetArea, self.players_dock)
        self.observers_dock = QDockWidget(GameSessionWindow)
        self.observers_dock.setObjectName(u"observers_dock")
        self.observers_dock.setFeatures(QDockWidget.DockWidgetFloatable|QDockWidget.DockWidgetMovable)
        self.observers_dock_contents = QWidget()
        self.observers_dock_contents.setObjectName(u"observers_dock_contents")
        self.observer_layout = QGridLayout(self.observers_dock_contents)
        self.observer_layout.setSpacing(6)
        self.observer_layout.setContentsMargins(11, 11, 11, 11)
        self.observer_layout.setObjectName(u"observer_layout")
        self.observers_dock.setWidget(self.observers_dock_contents)
        GameSessionWindow.addDockWidget(Qt.TopDockWidgetArea, self.observers_dock)
        self.game_dock = QDockWidget(GameSessionWindow)
        self.game_dock.setObjectName(u"game_dock")
        self.game_dock.setFeatures(QDockWidget.DockWidgetFloatable|QDockWidget.DockWidgetMovable)
        self.game_dock_contents = QWidget()
        self.game_dock_contents.setObjectName(u"game_dock_contents")
        self.game_dock_layout = QGridLayout(self.game_dock_contents)
        self.game_dock_layout.setSpacing(6)
        self.game_dock_layout.setContentsMargins(11, 11, 11, 11)
        self.game_dock_layout.setObjectName(u"game_dock_layout")
        self.generate_game_label = QLabel(self.game_dock_contents)
        self.generate_game_label.setObjectName(u"generate_game_label")
        self.generate_game_label.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.game_dock_layout.addWidget(self.generate_game_label, 0, 0, 1, 2)

        self.view_game_details_button = QPushButton(self.game_dock_contents)
        self.view_game_details_button.setObjectName(u"view_game_details_button")

        self.game_dock_layout.addWidget(self.view_game_details_button, 0, 2, 1, 1)

        self.save_iso_button = QToolButton(self.game_dock_contents)
        self.save_iso_button.setObjectName(u"save_iso_button")
        self.save_iso_button.setMinimumSize(QSize(100, 0))
        self.save_iso_button.setPopupMode(QToolButton.MenuButtonPopup)

        self.game_dock_layout.addWidget(self.save_iso_button, 0, 3, 1, 1)

        self.customize_user_preferences_button = QPushButton(self.game_dock_contents)
        self.customize_user_preferences_button.setObjectName(u"customize_user_preferences_button")

        self.game_dock_layout.addWidget(self.customize_user_preferences_button, 0, 4, 1, 1)

        self.line_generate = QFrame(self.game_dock_contents)
        self.line_generate.setObjectName(u"line_generate")
        self.line_generate.setFrameShape(QFrame.HLine)
        self.line_generate.setFrameShadow(QFrame.Sunken)

        self.game_dock_layout.addWidget(self.line_generate, 1, 0, 1, 5)

        self.background_process_button = QToolButton(self.game_dock_contents)
        self.background_process_button.setObjectName(u"background_process_button")
        self.background_process_button.setMinimumSize(QSize(140, 0))
        self.background_process_button.setPopupMode(QToolButton.MenuButtonPopup)

        self.game_dock_layout.addWidget(self.background_process_button, 2, 0, 1, 1)

        self.progress_label = QLabel(self.game_dock_contents)
        self.progress_label.setObjectName(u"progress_label")
        sizePolicy3 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.progress_label.sizePolicy().hasHeightForWidth())
        self.progress_label.setSizePolicy(sizePolicy3)
        self.progress_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.progress_label.setWordWrap(True)

        self.game_dock_layout.addWidget(self.progress_label, 2, 1, 1, 2)

        self.progress_bar = QProgressBar(self.game_dock_contents)
        self.progress_bar.setObjectName(u"progress_bar")
        self.progress_bar.setValue(0)
        self.progress_bar.setInvertedAppearance(False)

        self.game_dock_layout.addWidget(self.progress_bar, 2, 3, 1, 2)

        self.game_dock.setWidget(self.game_dock_contents)
        GameSessionWindow.addDockWidget(Qt.TopDockWidgetArea, self.game_dock)
        self.history_dock = QDockWidget(GameSessionWindow)
        self.history_dock.setObjectName(u"history_dock")
        self.history_dock.setFeatures(QDockWidget.DockWidgetFloatable|QDockWidget.DockWidgetMovable)
        self.history_dock_contents = QWidget()
        self.history_dock_contents.setObjectName(u"history_dock_contents")
        self.history_layout = QVBoxLayout(self.history_dock_contents)
        self.history_layout.setSpacing(6)
        self.history_layout.setContentsMargins(11, 11, 11, 11)
        self.history_layout.setObjectName(u"history_layout")
        self.history_layout.setContentsMargins(0, 0, 0, 0)
        self.history_table_widget = QTableWidget(self.history_dock_contents)
        if (self.history_table_widget.columnCount() < 5):
            self.history_table_widget.setColumnCount(5)
        __qtablewidgetitem = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(3, __qtablewidgetitem3)
        __qtablewidgetitem4 = QTableWidgetItem()
        self.history_table_widget.setHorizontalHeaderItem(4, __qtablewidgetitem4)
        self.history_table_widget.setObjectName(u"history_table_widget")
        self.history_table_widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.history_table_widget.setSortingEnabled(True)
        self.history_table_widget.horizontalHeader().setVisible(False)
        self.history_table_widget.horizontalHeader().setDefaultSectionSize(200)
        self.history_table_widget.verticalHeader().setVisible(False)

        self.history_layout.addWidget(self.history_table_widget)

        self.history_dock.setWidget(self.history_dock_contents)
        GameSessionWindow.addDockWidget(Qt.TopDockWidgetArea, self.history_dock)
        self.audit_dock = QDockWidget(GameSessionWindow)
        self.audit_dock.setObjectName(u"audit_dock")
        self.audit_dock.setFeatures(QDockWidget.DockWidgetFloatable|QDockWidget.DockWidgetMovable)
        self.audit_dock_contents = QWidget()
        self.audit_dock_contents.setObjectName(u"audit_dock_contents")
        self.audit_dock_layout = QVBoxLayout(self.audit_dock_contents)
        self.audit_dock_layout.setSpacing(0)
        self.audit_dock_layout.setContentsMargins(11, 11, 11, 11)
        self.audit_dock_layout.setObjectName(u"audit_dock_layout")
        self.audit_dock_layout.setContentsMargins(0, 2, 0, 0)
        self.audit_table_widget = QTableWidget(self.audit_dock_contents)
        if (self.audit_table_widget.columnCount() < 3):
            self.audit_table_widget.setColumnCount(3)
        __qtablewidgetitem5 = QTableWidgetItem()
        self.audit_table_widget.setHorizontalHeaderItem(0, __qtablewidgetitem5)
        __qtablewidgetitem6 = QTableWidgetItem()
        self.audit_table_widget.setHorizontalHeaderItem(1, __qtablewidgetitem6)
        __qtablewidgetitem7 = QTableWidgetItem()
        self.audit_table_widget.setHorizontalHeaderItem(2, __qtablewidgetitem7)
        self.audit_table_widget.setObjectName(u"audit_table_widget")
        self.audit_table_widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.audit_table_widget.setSortingEnabled(True)
        self.audit_table_widget.horizontalHeader().setVisible(True)
        self.audit_table_widget.horizontalHeader().setDefaultSectionSize(200)
        self.audit_table_widget.verticalHeader().setVisible(False)

        self.audit_dock_layout.addWidget(self.audit_table_widget)

        self.audit_dock.setWidget(self.audit_dock_contents)
        GameSessionWindow.addDockWidget(Qt.LeftDockWidgetArea, self.audit_dock)

        self.retranslateUi(GameSessionWindow)

        QMetaObject.connectSlotsByName(GameSessionWindow)
    # setupUi

    def retranslateUi(self, GameSessionWindow):
        GameSessionWindow.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"Game Session", None))
        self.action_add_player.setText(QCoreApplication.translate("GameSessionWindow", u"Add player", None))
        self.action_add_row.setText(QCoreApplication.translate("GameSessionWindow", u"Add row", None))
        self.rename_session_action.setText(QCoreApplication.translate("GameSessionWindow", u"Change title", None))
        self.change_password_action.setText(QCoreApplication.translate("GameSessionWindow", u"Change password", None))
        self.delete_session_action.setText(QCoreApplication.translate("GameSessionWindow", u"Delete session", None))
        self.actionbar.setText(QCoreApplication.translate("GameSessionWindow", u"bar", None))
        self.actionasdf.setText(QCoreApplication.translate("GameSessionWindow", u"asdf", None))
        self.game_connection_tool.setText(QCoreApplication.translate("GameSessionWindow", u"Connect to game", None))
        self.game_connection_label.setText(QCoreApplication.translate("GameSessionWindow", u"Game: Disconnected", None))
        self.server_connection_button.setText(QCoreApplication.translate("GameSessionWindow", u"Connect", None))
        self.server_connection_label.setText(QCoreApplication.translate("GameSessionWindow", u"Server: Disconnected", None))
        self.session_status_tool.setText(QCoreApplication.translate("GameSessionWindow", u"Start", None))
        self.session_status_label.setText(QCoreApplication.translate("GameSessionWindow", u"Session: Not Started", None))
        self.advanced_options_tool.setText(QCoreApplication.translate("GameSessionWindow", u"Advanced options...", None))
        self.players_dock.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"Session", None))
        self.title_connection_state_label.setText(QCoreApplication.translate("GameSessionWindow", u"Connection state", None))
        self.new_row_button.setText(QCoreApplication.translate("GameSessionWindow", u"New Row", None))
        self.title_player_name_label.setText(QCoreApplication.translate("GameSessionWindow", u"Players", None))
        self.observers_dock.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"Observers", None))
        self.game_dock.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"Game", None))
        self.generate_game_label.setText(QCoreApplication.translate("GameSessionWindow", u"<Game not generated>", None))
        self.view_game_details_button.setText(QCoreApplication.translate("GameSessionWindow", u"View Spoiler", None))
        self.save_iso_button.setText(QCoreApplication.translate("GameSessionWindow", u"Save ISO", None))
        self.customize_user_preferences_button.setText(QCoreApplication.translate("GameSessionWindow", u"Customize in-game settings", None))
        self.background_process_button.setText(QCoreApplication.translate("GameSessionWindow", u"Stop", None))
        self.progress_label.setText("")
        self.history_dock.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"History", None))
        ___qtablewidgetitem = self.history_table_widget.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("GameSessionWindow", u"Provider", None));
        ___qtablewidgetitem1 = self.history_table_widget.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("GameSessionWindow", u"Receiver", None));
        ___qtablewidgetitem2 = self.history_table_widget.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("GameSessionWindow", u"Pickup", None));
        ___qtablewidgetitem3 = self.history_table_widget.horizontalHeaderItem(3)
        ___qtablewidgetitem3.setText(QCoreApplication.translate("GameSessionWindow", u"Location", None));
        ___qtablewidgetitem4 = self.history_table_widget.horizontalHeaderItem(4)
        ___qtablewidgetitem4.setText(QCoreApplication.translate("GameSessionWindow", u"Time", None));
        self.audit_dock.setWindowTitle(QCoreApplication.translate("GameSessionWindow", u"Audit Log", None))
        ___qtablewidgetitem5 = self.audit_table_widget.horizontalHeaderItem(0)
        ___qtablewidgetitem5.setText(QCoreApplication.translate("GameSessionWindow", u"User", None));
        ___qtablewidgetitem6 = self.audit_table_widget.horizontalHeaderItem(1)
        ___qtablewidgetitem6.setText(QCoreApplication.translate("GameSessionWindow", u"Message", None));
        ___qtablewidgetitem7 = self.audit_table_widget.horizontalHeaderItem(2)
        ___qtablewidgetitem7.setText(QCoreApplication.translate("GameSessionWindow", u"Time", None));
    # retranslateUi




import copy
from collections import defaultdict
from random import Random
from typing import List, Dict, Optional, Tuple

from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.game_patches import ElevatorConnection
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.game_description.world.world_list import WorldList


class ElevatorHelper:
    teleporter: Teleporter
    destination: AreaLocation
    connected_elevator: Optional["ElevatorHelper"]

    def __init__(self, teleporter: Teleporter, destination: AreaLocation):
        self.teleporter = teleporter
        self.destination = destination
        self.connected_elevator = None

    @property
    def world_asset_id(self):
        return self.teleporter.world_asset_id

    @property
    def area_asset_id(self):
        return self.teleporter.area_asset_id

    @property
    def instance_id(self):
        return self.teleporter.instance_id

    def connect_to(self, other: "ElevatorHelper"):
        self.destination = other.teleporter.area_location
        other.destination = self.teleporter.area_location
        self.connected_elevator = other
        other.connected_elevator = self

    @property
    def area_location(self):
        return AreaLocation(self.world_asset_id, self.area_asset_id)


def try_randomize_elevators(rng: Random,
                            echoes_elevators: Tuple[ElevatorHelper, ...],
                            ) -> List[ElevatorHelper]:
    elevator_database: List[ElevatorHelper] = list(echoes_elevators)
    assert len(elevator_database) % 2 == 0

    elevator_list = copy.copy(elevator_database)
    elevators_by_world: Dict[int, List[ElevatorHelper]] = defaultdict(list)
    for elevator in elevator_list:
        elevators_by_world[elevator.world_asset_id].append(elevator)

    while elevator_list:
        source_elevators: List[ElevatorHelper] = max(elevators_by_world.values(), key=len)
        target_elevators: List[ElevatorHelper] = [
            elevator
            for elevator in elevator_list
            if elevator not in source_elevators
        ]
        source_elevator = source_elevators[0]
        target_elevator = rng.choice(target_elevators)

        source_elevator.connect_to(target_elevator)

        elevators_by_world[source_elevator.world_asset_id].remove(source_elevator)
        elevators_by_world[target_elevator.world_asset_id].remove(target_elevator)
        elevator_list.remove(source_elevator)
        elevator_list.remove(target_elevator)

    # TODO
    list3 = copy.copy(elevator_database)
    celevator_list3 = [list3[0]]
    while list3:
        celevator_list1 = []
        for celevator1 in celevator_list3:
            index = 0
            while index < len(list3):
                celevator2 = list3[index]
                if (celevator2.world_asset_id == celevator1.world_asset_id
                        or celevator2.area_asset_id == celevator1.destination.area_asset_id):
                    celevator_list1.append(celevator2)
                    list3.remove(celevator2)
                else:
                    index += 1
        if celevator_list1:
            celevator_list3 = celevator_list1
        else:
            # Randomization failed
            return try_randomize_elevators(rng, echoes_elevators)

    return elevator_database


def two_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[ElevatorHelper, ...],
                                 between_areas: bool
                                 ) -> ElevatorConnection:
    if len(elevator_database) % 2 != 0:
        raise ValueError("Two-way elevator shuffle, but odd number of elevators to shuffle.")
    if between_areas:
        elevator_database = try_randomize_elevators(rng, elevator_database)
    else:
        elevators = list(elevator_database)
        rng.shuffle(elevators)
        while elevators:
            elevators.pop().connect_to(elevators.pop())

    return {
        elevator.teleporter: elevator.connected_elevator.area_location
        for elevator in elevator_database
    }


def one_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[ElevatorHelper, ...],
                                 target_locations: List[AreaLocation],
                                 replacement: bool,
                                 ) -> ElevatorConnection:
    target_locations.sort()
    rng.shuffle(target_locations)

    def _create_target():
        if replacement:
            return rng.choice(target_locations)
        else:
            return target_locations.pop()

    return {
        elevator.teleporter: _create_target()
        for elevator in elevator_database
    }


def create_elevator_database(world_list: WorldList,
                             all_teleporters: List[Teleporter],
                             ) -> Tuple[ElevatorHelper, ...]:
    """
    Creates a tuple of Elevator objects, exclude those that belongs to one of the areas provided.
    :param world_list:
    :param all_teleporters: Set of teleporters to use
    :return:
    """
    all_helpers = [
        ElevatorHelper(node.teleporter, node.default_connection)

        for world, area, node in world_list.all_worlds_areas_nodes
        if isinstance(node, TeleporterNode)
    ]
    return tuple(
        helper
        for helper in all_helpers
        if helper.teleporter in all_teleporters
    )



from mock import MagicMock, AsyncMock

from randovania.cli.commands import batch_distribute
from randovania.layout.permalink import Permalink


def test_batch_distribute_helper(mocker):
    # Setup
    description = MagicMock()
    mock_generate_description: AsyncMock = mocker.patch(
        "randovania.generator.generator.generate_and_validate_description",
        new_callable=AsyncMock, return_value=description)
    mock_perf_counter = mocker.patch("time.perf_counter", autospec=False)  # TODO: pytest-qt bug

    base_permalink = MagicMock()
    seed_number = 5000
    validate = MagicMock()
    output_dir = MagicMock()
    timeout = 67
    description.file_extension.return_value = "rdvgame"

    expected_permalink = Permalink(
        seed_number=seed_number,
        spoiler=True,
        presets=base_permalink.presets,
    )

    mock_perf_counter.side_effect = [1000, 5000]

    # Run
    delta_time = batch_distribute.batch_distribute_helper(base_permalink, seed_number, timeout, validate, output_dir)

    # Assert
    mock_generate_description.assert_awaited_once_with(permalink=expected_permalink, status_update=None,
                                                       validate_after_generation=validate, timeout=timeout,
                                                       attempts=0)

    assert delta_time == 4000
    output_dir.joinpath.assert_called_once_with("{}.rdvgame".format(seed_number))
    description.save_to_file.assert_called_once_with(output_dir.joinpath.return_value)



import dataclasses
import uuid

import pytest
from PySide2 import QtCore
from PySide2.QtCore import Qt

from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.gui.preset_settings.starting_area_tab import PresetStartingArea
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.base_configuration import StartingLocationList


@pytest.mark.parametrize("game", [RandovaniaGame.METROID_PRIME, RandovaniaGame.METROID_PRIME_ECHOES, RandovaniaGame.METROID_PRIME_CORRUPTION])
def test_on_preset_changed(skip_qtbot, preset_manager, game):
    # Setup
    base = preset_manager.default_preset_for_game(game).get_preset()
    preset = dataclasses.replace(base,
                                 uuid=uuid.UUID('b41fde84-1f57-4b79-8cd6-3e5a78077fa6'),
                                 base_preset_uuid=base.uuid)
    editor = PresetEditor(preset)
    window = PresetStartingArea(editor, default_database.game_description_for(preset.game))

    # Run
    window.on_preset_changed(editor.create_custom_preset_with())

    # Assert
    num_areas = len(StartingLocationList.areas_list(preset.game))
    assert len(window._starting_location_for_area) == num_areas


def test_starting_location_world_select(skip_qtbot, preset_manager):
    # Setup
    base = preset_manager.default_preset_for_game(RandovaniaGame.METROID_PRIME_ECHOES).get_preset()
    preset = dataclasses.replace(base,
                                 uuid=uuid.UUID('b41fde84-1f57-4b79-8cd6-3e5a78077fa6'),
                                 base_preset_uuid=base.uuid)
    editor = PresetEditor(preset)
    window = PresetStartingArea(editor, default_database.game_description_for(preset.game))
    skip_qtbot.addWidget(window)

    # Run
    checkbox_list = window._starting_location_for_world
    window.on_preset_changed(editor.create_custom_preset_with())
    assert len(checkbox_list) == 10
    temple_grounds_checkbox = checkbox_list["Temple Grounds"]
    assert temple_grounds_checkbox.checkState() == Qt.PartiallyChecked
    skip_qtbot.mouseClick(temple_grounds_checkbox, Qt.LeftButton)
    assert temple_grounds_checkbox.checkState() == Qt.Checked
    assert len(editor.configuration.starting_location.locations) == 39
    skip_qtbot.mouseClick(temple_grounds_checkbox, Qt.LeftButton)
    assert temple_grounds_checkbox.checkState() == Qt.Unchecked
    assert len(editor.configuration.starting_location.locations) == 0
    skip_qtbot.mouseClick(temple_grounds_checkbox, Qt.LeftButton)
    window.on_preset_changed(editor.create_custom_preset_with())
    assert temple_grounds_checkbox.checkState() == Qt.Checked
    assert len(editor.configuration.starting_location.locations) == 39


@pytest.mark.parametrize("game_enum", RandovaniaGame)
def test_quick_fill_ship(skip_qtbot, preset_manager, game_enum: RandovaniaGame):
    # Setup
    base = preset_manager.default_preset_for_game(game_enum).get_preset()
    preset = dataclasses.replace(base,
                                 uuid=uuid.UUID('b41fde84-1f57-4b79-8cd6-3e5a78077fa6'),
                                 base_preset_uuid=base.uuid)
    editor = PresetEditor(preset)
    window = PresetStartingArea(editor, default_database.game_description_for(preset.game))
    skip_qtbot.addWidget(window)

    # Run
    skip_qtbot.mouseClick(window.starting_area_quick_fill_ship, QtCore.Qt.LeftButton)

    # Assert
    assert editor.configuration.starting_location.locations == frozenset([
        window.game_description.starting_location
    ])



from PySide2.QtCore import Qt
from PySide2.QtWidgets import QGridLayout, QLayout, QToolButton, QVBoxLayout, QWidget, QFrame, QScrollArea, QSizePolicy

class Foldable(QWidget):
    _mainLayout: QGridLayout
    _toggleButton: QToolButton
    _headerLine: QFrame
    _contentArea: QScrollArea
    _folded: bool

    def __init__(self, title: str, initially_folded: bool = True, parent: QWidget = None):
        super().__init__(parent)

        self._folded = initially_folded

        self._toggleButton = QToolButton(self)
        self._toggleButton.setStyleSheet("QToolButton { height: 20px; }")
        font = self._toggleButton.font()
        font.setBold(True)
        font.setPixelSize(13)
        self._toggleButton.setFont(font)
        self._toggleButton.setMaximumHeight(20)
        self._toggleButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self._toggleButton.setArrowType(Qt.ArrowType.RightArrow)
        self._toggleButton.setText(title)
        self._toggleButton.setCheckable(True)
        self._toggleButton.setChecked(False)
        self._toggleButton.clicked.connect(self._on_click)

        self._headerLine = QFrame(self)
        self._headerLine.setFrameShape(QFrame.HLine)
        self._headerLine.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        self._contentArea = QFrame(self)
        self._contentArea.setObjectName("foldable_contentArea")
        self._contentArea.setStyleSheet("#foldable_contentArea { border: none; }")
        self._contentArea.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)

        self._mainLayout = QGridLayout(self)
        self._mainLayout.setVerticalSpacing(0)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)
        self._mainLayout.addWidget(self._toggleButton, 0, 0, 1, 1, Qt.AlignLeft)
        self._mainLayout.addWidget(self._headerLine, 0, 1, 1, 1)
        self._mainLayout.addWidget(self._contentArea, 1, 0, 1, 2)

    @property
    def contents(self):
        return self._contentArea

    def _on_click(self, checked: bool):
        if self._folded:
            self._unfold()   
        else:
            self._fold()

    def _unfold(self):
        self._folded = False
        self._contentArea.show()
        self._toggleButton.setArrowType(Qt.ArrowType.DownArrow)

    def _fold(self):
        self._folded = True
        self._contentArea.hide()
        self._toggleButton.setArrowType(Qt.ArrowType.RightArrow)

    def set_content_layout(self, content_layout: QLayout):
        self._contentArea.setLayout(content_layout)
        if self._folded:
            self._fold()
        else:
            self._unfold()
